package nodeinfo

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
)

type fetcher struct {
	client   *http.Client
	matrix   bool
	nodeinfo bool
}

type Fetcher interface {
	QueryNode(domain string) ([]*Node, error)
}

func NewFetcher(opts ...QueryNodeOption) (Fetcher, error) {
	f := &fetcher{
		// don't follow redirects
		client: &http.Client{
			CheckRedirect: func(req *http.Request, via []*http.Request) error {
				return http.ErrUseLastResponse
			},
		},
		matrix:   true,
		nodeinfo: true,
	}

	for _, fn := range opts {
		if err := fn(f); err != nil {
			return nil, fmt.Errorf("could not apply option: %w", err)
		}
	}
	return f, nil
}

func (f *fetcher) validate(domain string) (string, error) {
	// validate options
	if !(f.nodeinfo || f.matrix) {
		return "", fmt.Errorf("all supported discovery protocols are disabled")
	}

	// validate user input
	if !(strings.HasPrefix(domain, "http://") || strings.HasPrefix(domain, "https://")) {
		domain = "https://" + domain
	}
	urlDomain, err := url.Parse(domain)
	if err != nil {
		return "", fmt.Errorf("parsing domain failed: %w", err)
	}

	return urlDomain.Host, nil
}

func (f *fetcher) QueryNode(rawDomain string) ([]*Node, error) {
	domain, err := f.validate(rawDomain)
	if err != nil {
		return nil, fmt.Errorf("validation of fetcher failed: %w", err)
	}

	var el []error
	var infos []*Node

	if f.nodeinfo {
		n, found, err := f.doNodeInfo(domain)
		if found && n != nil {
			infos = append(infos, n)
		}
		el = append(el, err)

		if !found {
			n, found, err = f.doNodeInfoX(domain)
			if found && n != nil {
				infos = append(infos, n)
			}
			el = append(el, err)
		}
	}

	if f.matrix {
		n, found, err := f.doMatrix(domain)
		if found && n != nil {
			infos = append(infos, n)
		}
		el = append(el, err)
	}

	if len(infos) != 0 {
		return infos, nil
	}

	return nil, fmt.Errorf("could not detect well-known service: %w", errors.Join(el...))
}

type nodeInfoLookup struct {
	Links []struct {
		Schema string `json:"rel"`
		Href   string `json:"href"`
	} `json:"links"`
}

type nodeInfoV1 struct {
	Software          Software               `json:"software"`
	Version           string                 `json:"version"`
	OpenRegistrations bool                   `json:"openRegistrations"`
	Protocols         Services               `json:"protocols"`
	Services          Services               `json:"services"`
	Usage             Usage                  `json:"usage"`
	Metadata          map[string]interface{} `json:"metadata"`
}

func (n *nodeInfoV1) into() *Node {
	if n == nil {
		return nil
	}
	protocolMap := make(map[string]struct{})
	for _, v := range append(n.Protocols.Inbound, n.Protocols.Outbound...) {
		protocolMap[v] = struct{}{}
	}
	var protocolList []string
	for k := range protocolMap {
		protocolList = append(protocolList, k)
	}
	return &Node{
		Version:           n.Version,
		Software:          n.Software,
		OpenRegistrations: n.OpenRegistrations,
		Protocols:         protocolList,
		Services:          n.Services,
		Usage:             n.Usage,
		Metadata:          n.Metadata,
	}
}

func (f *fetcher) doNodeInfo(domain string) (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/nodeinfo", schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get nodeinfo endpoints
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := nodeInfoLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	links := make(map[string]string)
	for _, link := range lookup.Links {
		// expect http and https
		links[strings.Replace(link.Schema, "https://", "http://", 1)] = link.Href
	}
	if len(links) == 0 {
		return nil, true, fmt.Errorf("broken lookup endpoint, no links found")
	}

	var href string
	isV1 := false
	// v2.2 is not jet officially published but already used in the wild ;)
	if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.2"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.1"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.0"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.1"]; ok {
		isV1 = true
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.0"]; ok {
		isV1 = true
		href = ref
	} else {
		return nil, true, fmt.Errorf("broken lookup endpoint, unknown schema(s) found: %s", links)
	}

	resp, err = f.client.Get(href)
	if err != nil {
		return nil, false, err
	}
	rawResp, err = io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}

	node := &Node{}
	if !isV1 {
		if err := json.Unmarshal(rawResp, node); err != nil {
			return nil, true, err
		}
	} else {
		v1Node := &nodeInfoV1{}
		if err := json.Unmarshal(rawResp, v1Node); err != nil {
			return nil, true, err
		}
		node = v1Node.into()
	}

	if err := Extend(f.client, domain, node); err != nil {
		return node, true, err
	}

	return node, true, nil
}

// nodeInfoX as proposed by https://github.com/jaywink/nodeinfo2/blob/master/schemas/1.0/schema.json
type nodeInfoX struct {
	Server struct {
		Software string `json:"software"`
		Version  string `json:"version"`
		Name     string `json:"name"`
		BaseURL  string `json:"baseUrl"`
	} `json:"server"`
	Version           string   `json:"version"`
	OpenRegistrations bool     `json:"openRegistrations"`
	Protocols         []string `json:"protocols"`
	Services          Services `json:"services"`
	Usage             struct {
		Users struct {
			Total          int64 `json:"total"`
			ActiveHalfyear int64 `json:"activeHalfyear"`
			ActiveMonth    int64 `json:"activeMonth"`
			ActiveWeek     int64 `json:"activeWeek"`
		} `json:"users"`
	} `json:"usage"`
	Relay        string `json:"relay"`
	Organization struct {
		Name    string `json:"name"`
		Contact string `json:"contact"`
		Account string `json:"account"`
	} `json:"organization"`
}

func (node *nodeInfoX) into() *Node {
	if node == nil {
		return nil
	}
	return &Node{
		Version: "2.2",
		Instance: Instance{
			Name: node.Server.Name,
		},
		Software: Software{
			Name:    node.Server.Software,
			Version: node.Server.Version,
		},
		OpenRegistrations: node.OpenRegistrations,
		Protocols:         node.Protocols,
		Services:          node.Services,
		Usage: Usage{
			Users: UsersUsage{
				Total:          node.Usage.Users.Total,
				ActiveHalfyear: node.Usage.Users.ActiveHalfyear,
				ActiveMonth:    node.Usage.Users.ActiveMonth,
				ActiveWeek:     node.Usage.Users.ActiveWeek,
			},
		},
		Metadata: map[string]interface{}{
			"relay": node.Relay,
			"nodeAdmins": []struct{ Name, Email, Account string }{{
				Name:    node.Organization.Name,
				Email:   node.Organization.Contact,
				Account: node.Organization.Account,
			}},
			"nativeNodeinfoVersion": fmt.Sprintf(`"x-nodeinfo2": "%s"`, node.Version),
		},
	}
}

func (f *fetcher) doNodeInfoX(domain string) (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/x-nodeinfo2", schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no x-nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	node := &nodeInfoX{}
	if err := json.Unmarshal(rawResp, node); err != nil {
		return nil, true, err
	}

	return node.into(), true, nil
}

func (f *fetcher) jsonClient(url string, obj any) (found bool, err error) {
	resp, err := f.client.Get(url)
	if err != nil {
		return false, err
	}
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return true, err
	}
	return true, json.Unmarshal(rawResp, obj)
}
