package nodeinfo

import (
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"strings"
)

// list from https://matrix.org/ecosystem/servers/
var mapMatrixSoftware = map[string]struct {
	www string
	git string
}{
	"synapse":   {www: "https://element-hq.github.io/synapse/latest/", git: "https://github.com/element-hq/synapse"},
	"dendrite":  {www: "https://matrix-org.github.io/dendrite/", git: "https://github.com/matrix-org/dendrite"},
	"conduit":   {www: "https://conduit.rs/", git: "https://gitlab.com/famedly/conduit"},
	"construct": {git: "https://github.com/matrix-construct/construct"},
}

const (
	wellKnownServer     = "%s://%s/.well-known/matrix/server"
	wellKnownSupport    = "%s://%s/.well-known/matrix/support"
	matrixServerVersion = "%s://%s/_matrix/federation/v1/version"
)

type matrixLookup struct {
	ServerAddr string `json:"m.server"`
}

type matrixVersion struct {
	Server struct {
		ImplName    string `json:"name"`
		ImplVersion string `json:"version"`
	} `json:"server"`
}

type matrixSupport struct {
	SupportPage string `json:"support_page"`
	Contacts    []struct {
		Role     string `json:"role"`
		Email    string `json:"email_address"`
		MatrixID string `json:"matrix_id"`
	} `json:"contacts"`
}

type matrixAdmin struct {
	Email   string `json:"email,omitempty"`
	Account string `json:"account,omitempty"`
	Role    string `json:"role,omitempty"`
}

func (f *fetcher) doMatrix(domain string) (*Node, bool, error) {
	schema := "https"
	node := Node{
		Version:   "2.2",
		Protocols: []string{ProtocolMatrix},
		Metadata: map[string]interface{}{
			"nativeNodeinfoVersion": `"matrix": ""`,
		},
	}

	// TODO: respect DNS lookup as per https://spec.matrix.org/v1.8/server-server-api/#resolving-server-names

	resp, err := f.client.Get(fmt.Sprintf(wellKnownServer, schema, domain))
	if err != nil {
		return nil, false, err
	}

	// no matrix server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get matrix server address
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := matrixLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	// get server software
	version := matrixVersion{}
	found, err := f.jsonClient(fmt.Sprintf(matrixServerVersion, schema, lookup.ServerAddr), &version)
	if err != nil {
		return nil, found, err
	}
	node.Software.Version = version.Server.ImplVersion
	node.Software.Name = version.Server.ImplName
	if soft, found := mapMatrixSoftware[strings.ToLower(version.Server.ImplName)]; found {
		node.Software.Homepage = soft.www
		node.Software.Repository = soft.git
	}

	// fetch support info as per MSC1929 (https://github.com/matrix-org/matrix-spec/pull/1733)
	support := matrixSupport{}
	if _, err := f.jsonClient(fmt.Sprintf(wellKnownSupport, schema, domain), &support); err == nil {
		if support.SupportPage != "" {
			node.Metadata["matrixSupportPage"] = support.SupportPage
		}
		admins := []matrixAdmin{}
		for _, contact := range support.Contacts {
			admins = append(admins, matrixAdmin{
				Email:   contact.Email,
				Account: contact.MatrixID,
				Role:    contact.Role,
			})
		}
		node.Metadata["matrixContacts"] = admins
	}

	// TODO: old trick (https://codeberg.org/thefederationinfo/py2-federation/src/branch/master/federation/hostmeta/parsers.py#L141-L144) don't work anymore
	node.OpenRegistrations = false

	return &node, true, nil
}
