package nodeinfo_test

import (
	"encoding/json"
	"testing"

	"github.com/jarcoal/httpmock"
	"github.com/stretchr/testify/assert"
	"github.com/tidwall/pretty"

	"codeberg.org/thefederationinfo/nodeinfo-go"
	_ "codeberg.org/thefederationinfo/nodeinfo-go/extension"
	known_extension "codeberg.org/thefederationinfo/nodeinfo_metadata_survey"
)

var useMock bool = true

func TestFetcherDoMatrix(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://matrix.org/.well-known/nodeinfo",
			httpmock.NewStringResponder(404, "<!doctype html><html lang=\"en\"><head><meta charset=\"utf-8\" /><meta property=\"og:title\" content=\"Matrix.org\" /><meta property=\"og:image\" content=\"https://matrix.org/blog/img/matrix-logo.png\" /><title>Matrix.org - Not Found</title><link rel=\"icon\" type=\"image/svg+xml\" href=\"/assets/favicon.svg\" /></head><body><main><div class=\"content\"><h1>Not Found</h1><p>This page doesn't exist.</p></div></main><p><a href=\"/legal/copyright-notice\">Copyright Notice</a></p></footer></body></html>"))
		httpmock.RegisterResponder("GET", "https://matrix.org/.well-known/matrix/server",
			httpmock.NewStringResponder(200, `{ "m.server": "matrix-federation.matrix.org:443" }`))
		httpmock.RegisterResponder("GET", "https://matrix-federation.matrix.org:443/_matrix/federation/v1/version",
			httpmock.NewStringResponder(200, `{"server":{"name":"Synapse","version":"1.101.0 (b=matrix-org-hotfixes,beff0a756e)"}}`))
		httpmock.RegisterResponder("GET", "https://matrix.org/.well-known/matrix/support",
			httpmock.NewStringResponder(200, `{"support_page": "https://matrix.org/contact/", "contacts": [{"role": "m.role.admin", "email_address": "abuse@matrix.org"}, {"role": "m.role.security", "email_address": "security@matrix.org"}]}`))
	}

	infos, err := f.QueryNode("matrix.org")
	assert.NoError(t, err)
	if assert.NotNil(t, infos) {
		info := infos[0]
		metadata, err := json.Marshal(info.Metadata)
		metadata = pretty.Pretty(metadata)
		assert.NoError(t, err)
		info.Metadata = nil
		assert.EqualValues(t, &nodeinfo.Node{
			Version:  "2.2",
			Instance: nodeinfo.Instance{},
			Software: nodeinfo.Software{
				Name:       "Synapse",
				Version:    "1.101.0 (b=matrix-org-hotfixes,beff0a756e)",
				Repository: "https://github.com/element-hq/synapse",
				Homepage:   "https://element-hq.github.io/synapse/latest/",
			},
			Protocols: []string{"matrix"},
		}, info)
		assert.EqualValues(t, `{
  "matrixContacts": [
    {
      "email": "abuse@matrix.org",
      "role": "m.role.admin"
    },
    {
      "email": "security@matrix.org",
      "role": "m.role.security"
    }
  ],
  "matrixSupportPage": "https://matrix.org/contact/",
  "nativeNodeinfoVersion": "\"matrix\": \"\""
}
`, string(metadata))
	}
}

const mastodonNodeInfo = `{
  "version": "2.0",
  "software": {
    "name": "mastodon",
    "version": "4.1.7"
  },
  "protocols": [
    "activitypub"
  ],
  "services": {
    "outbound": [],
    "inbound": []
  },
  "usage": {
    "users": {
      "total": 11290,
      "activeMonth": 6076,
      "activeHalfyear": 7249
    },
    "localPosts": 4925353
  },
  "openRegistrations": false,
  "metadata": {}
}`

func TestFetcherDoMastodon(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://chaos.social/.well-known/nodeinfo",
			httpmock.NewStringResponder(200, `{"links":[{"rel":"http://nodeinfo.diaspora.software/ns/schema/2.0","href":"https://chaos.social/nodeinfo/2.0"}]}`))
		httpmock.RegisterResponder("GET", "https://chaos.social/nodeinfo/2.0",
			httpmock.NewStringResponder(200, mastodonNodeInfo))
	}

	infos, err := f.QueryNode("chaos.social")
	assert.NoError(t, err)
	if assert.NotNil(t, infos) {
		info := infos[0]
		assert.EqualValues(t, "mastodon", info.Software.Name)
		assert.Truef(t, len(info.Software.Version) != 0, "expect a version")
		assert.EqualValues(t, []string{"activitypub"}, info.Protocols)
	}
}

const diasporaNodeInfo = `{
  "version": "2.1",
  "software": {
    "name": "diaspora",
    "version": "0.7.18.2-p6da4038c",
    "repository": "https://github.com/diaspora/diaspora",
    "homepage": "https://diasporafoundation.org/"
  },
  "protocols": [
    "diaspora"
  ],
  "services": {
    "inbound": [],
    "outbound": []
  },
  "openRegistrations": true,
  "usage": {
    "users": {
      "total": 105727,
      "activeHalfyear": 1525,
      "activeMonth": 596
    },
    "localPosts": 901765,
    "localComments": 970058
  },
  "metadata": {
    "nodeName": "diaspora",
    "xmppChat": false,
    "camo": {
      "markdown": true,
      "opengraph": true,
      "remotePods": true
    },
    "adminAccount": "davidmorley"
  }
}`

func TestFetcherDoDiaspora(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://diasp.org/.well-known/nodeinfo",
			httpmock.NewStringResponder(200, `{"links":[{"rel":"http://nodeinfo.diaspora.software/ns/schema/1.0","href":"https://diasp.org/nodeinfo/1.0"},{"rel":"http://nodeinfo.diaspora.software/ns/schema/2.0","href":"https://diasp.org/nodeinfo/2.0"},{"rel":"http://nodeinfo.diaspora.software/ns/schema/2.1","href":"https://diasp.org/nodeinfo/2.1"}]}`))
		httpmock.RegisterResponder("GET", "https://diasp.org/nodeinfo/2.1",
			httpmock.NewStringResponder(200, diasporaNodeInfo))
	}

	infos, err := f.QueryNode("diasp.org")
	assert.NoError(t, err)
	if assert.NotNil(t, infos) {
		info := infos[0]
		assert.EqualValues(t, "diaspora", info.Software.Name)
		assert.Truef(t, len(info.Software.Version) != 0, "expect a version")
		assert.EqualValues(t, []string{"diaspora"}, info.Protocols)
		assert.NotNil(t, info.Metadata["camo"])
		assert.Nil(t, info.Metadata["random"])
	}
}

func TestFetcherDoObermui(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://obermui.de/.well-known/nodeinfo",
			httpmock.NewStringResponder(307, "<html>\n<head><title>307 Temporary Redirect</title></head>\n<body>\n<center><h1>307 Temporary Redirect</h1></center>\n<hr><center>nginx</center>\n</body>\n</html>"))
		httpmock.RegisterResponder("GET", "https://obermui.de/.well-known/x-nodeinfo2",
			httpmock.NewStringResponder(404, `<html><head><title>not found</title></head></html>`))
		httpmock.RegisterResponder("GET", "https://obermui.de/.well-known/matrix/server",
			httpmock.NewStringResponder(200, `{"m.server": "obermui.de:443"}`))
		httpmock.RegisterResponder("GET", "https://obermui.de:443/_matrix/federation/v1/version",
			httpmock.NewStringResponder(200, `{"server":{"name":"Synapse","version":"1.90.0"}}`))
	}

	info, err := f.QueryNode("obermui.de")
	assert.NoError(t, err)
	assert.NotNil(t, info)
}

// https://lurk.org/.well-known/x-nodeinfo2
const prosodyNodeInfoX = `{
  "version": "1.0",
  "usage": {
    "users": {
      "total": 92,
      "activeHalfyear": 58,
      "activeMonth": 48,
      "activeWeek": 42
    }
  },
  "openRegistrations": true,
  "server": {
    "version": "0.11.9",
    "baseUrl": "https://lurk.org:5281/",
    "name": "lurk.org",
    "software": "Prosody"
  },
  "protocols": [
    "xmpp"
  ],
  "services": {
    "inbound": [
      "xmpp"
    ],
    "outbound": [
      "xmpp"
    ]
  }
}`

func TestFetcherDoProsody(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://lurk.org/.well-known/nodeinfo",
			httpmock.NewStringResponder(404, "<html>\n<head><title>404 Not Found</title></head>\n<body>\n<center><h1>404 Not Found</h1></center>\n<hr><center>nginx/1.18.0</center>\n</body>\n</html>"))
		httpmock.RegisterResponder("GET", "https://lurk.org/.well-known/x-nodeinfo2",
			httpmock.NewStringResponder(200, prosodyNodeInfoX))
	}

	infos, err := f.QueryNode("lurk.org")
	assert.NoError(t, err)
	if assert.NotNil(t, infos) {
		info := infos[0]
		assert.EqualValues(t, "Prosody", info.Software.Name)
		assert.EqualValues(t, []string{"xmpp"}, info.Protocols)
		assert.EqualValues(t, nodeinfo.Services{Inbound: []string{"xmpp"}, Outbound: []string{"xmpp"}}, info.Services)
	}
}

// https://xmpp.social/.well-known/x-nodeinfo2
const ejabberdNodeInfoX = `{
  "organization": {
    "name": "Hookipa admins",
    "contact": "xmpp:discussion@chat.hookipa.net?join"
  },
  "server": {
    "baseUrl": "https://xmpp.social/",
    "version": "23.01",
    "name": "xmpp.social",
    "software": "ejabberd"
  },
  "services": {
    "outbound": [
      "xmpp"
    ],
    "inbound": [
      "xmpp"
    ]
  },
  "protocols": [
    "xmpp"
  ],
  "version": "1.0",
  "openRegistrations": true,
  "usage": {
    "users": {
      "activeWeek":  261,
      "total": 921,
      "activeMonth": 345,
      "activeHalfyear": 641
    },
    "localPosts": 3680380,
    "localComments": 0
  }
}`

func TestFetcherDoEjabberd(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://xmpp.social/.well-known/nodeinfo",
			httpmock.NewStringResponder(404, "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n<html><head>\n<title>404 Not Found</title>\n</head><body>\n<h1>Not Found</h1>\n<p>The requested URL was not found on this server.</p>\n<hr>\n<address>Apache/2.4.57 (Debian) Server at xmpp.social Port 443</address>\n</body></html>"))
		httpmock.RegisterResponder("GET", "https://xmpp.social/.well-known/x-nodeinfo2",
			httpmock.NewStringResponder(200, ejabberdNodeInfoX))
	}

	infos, err := f.QueryNode("xmpp.social")
	assert.NoError(t, err)
	if assert.NotNil(t, infos) {
		info := infos[0]
		assert.EqualValues(t, "ejabberd", info.Software.Name)
		assert.EqualValues(t, []string{"xmpp"}, info.Protocols)
		assert.EqualValues(t, nodeinfo.Services{Inbound: []string{"xmpp"}, Outbound: []string{"xmpp"}}, info.Services)
	}
}

// https://socialhome.network/nodeinfo/1.0
const soclaHomeNodeInfo1 = `{
  "version": "1.0",
  "software": {
    "name": "socialhome",
    "version": "0.19.0"
  },
  "protocols": {
    "inbound": [
      "diaspora"
    ],
    "outbound": [
      "diaspora"
    ]
  },
  "services": {
    "inbound": [],
    "outbound": []
  },
  "openRegistrations": true,
  "usage": {
    "users": {
      "total": 1589,
      "activeHalfyear": 74,
      "activeMonth": 30
    },
    "localPosts": 34595,
    "localComments": 1514
  },
  "metadata": {
    "nodeName": "Socialhome"
  }
}`

// https://socialhome.network/.well-known/x-nodeinfo2
const soclaHomeNodeInfoX = `{
  "version": "1.0",
  "server": {
    "baseUrl": "https://socialhome.network",
    "name": "Socialhome",
    "software": "socialhome",
    "version": "0.19.0"
  },
  "organization": {
    "name": "TheAdmin",
    "contact": null,
    "account": "admin"
  },
  "protocols": [
    "diaspora"
  ],
  "relay": "none",
  "services": {
    "inbound": [],
    "outbound": []
  },
  "openRegistrations": true,
  "usage": {
    "users": {
      "total": 1589,
      "activeHalfyear": 74,
      "activeMonth": 30,
      "activeWeek": 14
    },
    "localPosts": 34595,
    "localComments": 1514
  }
}`

func TestFetcherDoSocialhomeNetwork(t *testing.T) {
	f, _ := nodeinfo.NewFetcher()
	if useMock {
		httpmock.Activate()
		defer httpmock.DeactivateAndReset()

		httpmock.RegisterResponder("GET", "https://socialhome.network/.well-known/nodeinfo",
			httpmock.NewStringResponder(200, `{"links": [{"rel": "http://nodeinfo.diaspora.software/ns/schema/1.0", "href": "https://socialhome.network/nodeinfo/1.0"}]}`))
		httpmock.RegisterResponder("GET", "https://socialhome.network/nodeinfo/1.0",
			httpmock.NewStringResponder(200, soclaHomeNodeInfo1))
		httpmock.RegisterResponder("GET", "https://socialhome.network/.well-known/x-nodeinfo2",
			httpmock.NewStringResponder(200, soclaHomeNodeInfoX))
	}

	infos, err := f.QueryNode("socialhome.network")
	assert.NoError(t, err)
	if assert.Len(t, infos, 1) {
		assert.EqualValues(t, &nodeinfo.Node{
			Version: "2.2",
			Instance: nodeinfo.Instance{
				Name:        "Socialhome",
				Description: "",
			},
			Software: nodeinfo.Software{
				Name:       "socialhome",
				Version:    "0.19.0",
				Repository: "",
				Homepage:   "",
			},
			OpenRegistrations: true,
			Protocols:         []string{"diaspora"},
			Services:          nodeinfo.Services{Inbound: []string{}, Outbound: []string{}},
			Usage: nodeinfo.Usage{
				Users: nodeinfo.UsersUsage{
					Total:          1589,
					ActiveHalfyear: 74,
					ActiveMonth:    30,
					ActiveWeek:     14,
				},
				LocalPosts:    34595,
				LocalComments: 1514,
			},
			Metadata: map[string]interface{}{
				"nodeName": "Socialhome",
				"relay":    "none",
				"nodeAdmins": []known_extension.NodeAdmin{{
					Name:    "TheAdmin",
					Account: "admin",
				}},
				"nativeNodeinfoVersion": "\"nodeinfo\": \"1.0\"",
			},
		},
			infos[0])
	}
}
